/*
 * Copyright (C) 2005-2010 Alfresco Software Limited.
 *
 * This file is part of Alfresco
 *
 * Alfresco is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Alfresco is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Alfresco. If not, see <http://www.gnu.org/licenses/>.
 */
package org.adullact.iparapheur.repo.office;

import org.adullact.iparapheur.repo.office.RuntimeExec.ExecutionResult;
import org.alfresco.error.AlfrescoRuntimeException;
import org.apache.log4j.Logger;
import org.springframework.context.ApplicationEvent;
import org.springframework.extensions.surf.util.AbstractLifecycleBean;

import java.util.Collections;
import java.util.List;
import java.util.concurrent.Semaphore;

/**
 * Application bootstrap bean that is able to execute one or more
 * native executable statements upon startup and shutdown.
 *
 * @author Derek Hulley
 * @author Lukas Hameury
 */
public class OfficeBootstrapAndRelaunch extends AbstractLifecycleBean
{
    private static Logger logger = Logger.getLogger(OfficeBootstrapAndRelaunch.class);

    private List<RuntimeExec> startupCommands;
    private boolean failOnError;
    private boolean enabled;
    private boolean hasToRelaunch;

    /** Keep track of the processes so that we can kill them on shutdown */
    private static Semaphore officeProcess;
    private static ExecutionResult executionResult;

    private Thread shutdownThread;
    private Thread relaunchThread;

    /**
     * Initializes the bean
     * <ul>
     *   <li>failOnError = true</li>
     *   <li>killProcessesOnShutdown = true</li>
     * </ul>
     */
    public OfficeBootstrapAndRelaunch()
    {
        this.startupCommands = Collections.emptyList();
        failOnError = true;
        enabled = true;
        hasToRelaunch = true;
    }

    /**
     * Set the commands to execute, in sequence, when the application context
     * is initialized.
     *
     * @param startupCommands list of commands
     */
    public void setStartupCommands(List<RuntimeExec> startupCommands)
    {
        this.startupCommands = startupCommands;
    }

    /**
     * Set whether a process failure generates an error or not.  Deviation from the default is
     * useful if use as part of a process where the command or the codes generated by the
     * execution may be ignored or avoided by the system.
     *
     * @param failOnError           <tt>true<tt> (default) to issue an error message and throw an
     *                              exception if the process fails to execute or generates an error
     *                              return value.
     *
     * @since 2.1
     */
    public void setFailOnError(boolean failOnError)
    {
        this.failOnError = failOnError;
    }


    public void setEnabled(String enabled)
    {
        this.enabled = Boolean.parseBoolean(enabled);
    }

    @Override
    protected synchronized void onBootstrap(ApplicationEvent event)
    {
        if(enabled) {
            officeProcess = new Semaphore(1);
            // execute
            launch();
            // Ajout LH : Launch relaunch thread
            relaunchThread = new CheckProcessAndRelaunchThread();
            relaunchThread.start();
            // Force a shutdown on VM termination as we can't rely on the Spring context termination
            this.shutdownThread = new KillProcessShutdownThread();
            Runtime.getRuntime().addShutdownHook(this.shutdownThread);
            // done
            if (logger.isDebugEnabled())
            {
                logger.debug("Bootstrap execution of " + startupCommands.size() + " commands was successful");
            }
        }
    }

    /**
     * A thread that serves to kill the successfully created process, if required
     *
     * @since 2.1
     * @author Derek Hulley
     */
    private class KillProcessShutdownThread extends Thread
    {
        @Override
        public void run()
        {
            try {
                doShutdown();
            } catch (InterruptedException e) {
                //Should not happen
                e.printStackTrace();
            }
        }
    }

    /**
     * Ajout LH
     * Thread permettant de checker le process, dès qu'il s'arrête, on le relance
     */
    private class CheckProcessAndRelaunchThread extends Thread
    {
        @Override
        public void run()
        {
            while(hasToRelaunch) {
                try {
                    executionResult.getProcess().waitFor();
                } catch (InterruptedException e) {
                    //Fin du check
                } finally {
                    if(hasToRelaunch) {
                        //Le service s'est arrêté, on le relance
                        logger.warn("Service Office arrêté\nLancement du service Office");
                        launch();
                    }
                }
            }
        }
    }

    /**
     * Handle the shutdown of a subsystem but not the entire VM
     */
    @Override
    protected synchronized void onShutdown(ApplicationEvent event)
    {
        if(enabled) {
            hasToRelaunch = false;
            try
            {
                doShutdown();
                // We managed to stop the process ourselves (e.g. on subsystem shutdown). Remove the shutdown hook
                Runtime.getRuntime().removeShutdownHook(this.shutdownThread);
            }
            catch (IllegalStateException e)
            {
                e.printStackTrace();
                // The system is shutting down - we'll have to let the shutdown hook run
            } catch (InterruptedException e) {
                e.printStackTrace();
                // The system is shutting down - we'll have to let the shutdown hook run
            }
        }
    }

    private void doShutdown() throws InterruptedException {
        if(executionResult != null) {
            officeProcess.acquire();
            executionResult.killProcess();
            officeProcess.release();
        }
    }

    public static void doRestart() throws InterruptedException {
        if(executionResult != null) {
            officeProcess.acquire();
            executionResult.killProcess();
            officeProcess.release();
        }
    }

    private void launch()
    {
        //Just in case... Should not happen
        try {
            doShutdown();
        } catch (InterruptedException e) {
            //Should not happen
            e.printStackTrace();
        }
        if(startupCommands.size() == 1) {
            ExecutionResult result = startupCommands.get(0).execute();
            // check for failure
            if (!result.getSuccess())
            {
                String msg = "Bootstrap command failed: \n" + result;
                if (failOnError)
                {
                    throw new AlfrescoRuntimeException(msg);
                }
                else
                {
                    logger.error(msg);
                }
            }
            else
            {
                // It executed, so keep track of it
                executionResult = result;
            }
        }
    }
}
